"""System testing of the whole package."""
from orangearg.argument.miner import reader, chunker, processor, miner

from .conftest import TEST_DATA_FOLDER


def test_system():
    """Systen testing of the whole package."""
    fpath = TEST_DATA_FOLDER / "reviews.json"

    # reader
    df_arguments = reader.read_json_file(fpath=fpath)
    df_arguments = df_arguments.dropna().reset_index(drop=True)
    df_arguments = df_arguments.rename(
        columns={"reviewText": "argument", "overall": "score"}
    )
    assert df_arguments["argument"].dtype == "object"
    assert df_arguments["score"].dtype == "int64"

    # chunker
    arguments = df_arguments["argument"]
    arg_scores = df_arguments["score"]
    chunk_arg_ids, chunks = chunker.get_chunk(docs=arguments)
    chunk_p_scores = chunker.get_chunk_polarity_score(chunks=chunks)
    chunk_topics, chunk_embeds, df_topics = chunker.get_chunk_topic(chunks=chunks)
    chunk_ranks = chunker.get_chunk_rank(arg_ids=chunk_arg_ids, embeds=chunk_embeds)
    assert all(-1 <= s <= 1 for s in chunk_p_scores)
    assert all(0 <= r <= 1 for r in chunk_ranks)
    assert (
        len(chunks)
        == len(chunk_arg_ids)
        == len(chunk_p_scores)
        == len(chunk_topics)
        == len(chunk_ranks)
        == chunk_embeds.shape[0]
    )

    df_chunks = chunker.get_chunk_table(
        arg_ids=chunk_arg_ids,
        chunks=chunks,
        p_scores=chunk_p_scores,
        topics=chunk_topics,
        ranks=chunk_ranks,
    )

    # processor
    arg_topics = processor.get_argument_topics(
        arg_ids=chunk_arg_ids, topics=chunk_topics
    )
    arg_sentiments = processor.get_argument_sentiment(
        arg_ids=chunk_arg_ids, ranks=chunk_ranks, p_scores=chunk_p_scores
    )
    arg_coherences = processor.get_argument_coherence(
        scores=arg_scores, sentiments=arg_sentiments
    )
    assert all(0 <= s <= 1 for s in arg_sentiments)
    assert all(0 < c <= 1 for c in arg_coherences)
    assert (
        len(arg_topics)
        == len(arg_sentiments)
        == len(arg_coherences)
        == df_arguments.shape[0]
    )

    df_arguments_processed = processor.update_argument_table(
        df_arguments=df_arguments,
        topics=arg_topics,
        sentiments=arg_sentiments,
        coherences=arg_coherences,
    )

    # miner
    last_topic = df_topics.iloc[-1]["topic"]
    arg_selection = miner.select_by_topic(data=df_arguments_processed, topic=last_topic)
    edges = miner.get_edges(data=arg_selection)
    weights = miner.get_edge_weights(data=arg_selection, edges=edges)
    assert all(i in arg_selection.index for edge in edges for i in edge)
    assert all(-1 <= w <= 1 for w in weights)

    df_edges = miner.get_edge_table(edges=edges, weights=weights)

    labels = miner.get_node_labels(
        indices=arg_selection.index.tolist(),
        sources=df_edges["source"].tolist(),
        targets=df_edges["target"].tolist(),
    )
    assert all(l in ["supportive", "defeated"] for l in labels)
    assert len(labels) == arg_selection.shape[0]

    df_nodes = miner.get_node_table(
        arg_ids=arg_selection["argument_id"].tolist(),
        arguments=arg_selection["argument"].tolist(),
        scores=arg_selection["score"].tolist(),
        labels=labels,
    )



"""Test configuration."""
from pathlib import Path
import json

import pytest
import pandas as pd

from orangearg.argument.miner.chunker import TopicModel

CWD = Path("__file__").absolute().parent
TEST_DATA_FOLDER = CWD / "tests" / "test_data"


@pytest.fixture(scope="function")
def df_chunks():
    """Dataframe of chunks.

    The dataframe contains 5 columns that are:
    - argument_id (int): id of arguments that the chunk belongs to
    - chunk (str): chunk text
    - topic (int): topic index of a chunk
    - rank (float): pagerank of a chunk within the argument it belongs to
    - polarity_score (float): sentiment polarity score of a chunk
    """
    fpath = TEST_DATA_FOLDER / "chunks.csv"
    return pd.read_csv(fpath, delimiter=";", index_col=0)


@pytest.fixture(scope="function")
def df_arguments():
    """Dataframe of arguments.

    This dataframe contains 2 columns that are:
    - argument (str): argument text
    - score (int): overall score alongside of an argument, in range of [1, 5].
    """
    fpath = TEST_DATA_FOLDER / "reviews.json"
    with open(fpath, "r", encoding="utf-8") as file:
        data = []
        for obj in file:
            data.append(json.loads(obj))
    result = pd.json_normalize(data)
    return result.rename(columns={"reviewText": "argument", "overall": "score"})


@pytest.fixture(scope="function")
def df_arguments_processed():
    """Dataframe of processed arguments.

    The dataframe contains 5 columns that are:
    - argument (str): the argument text
    - score (int): the overal score aligned with argument
    - topics (Tuple[int]): the topics mentioned by the argument
    - sentiment (float): the argument sentiment score
    - coherence (float): the argument coherence score
    """
    fpath = TEST_DATA_FOLDER / "arguments_processed.csv"
    return pd.read_csv(fpath, delimiter=";", index_col=0)


@pytest.fixture(scope="function")
def large_chunk_set(df_chunks):
    """List of around 1200 chunk texts."""
    return df_chunks["chunk"].dropna().tolist()


@pytest.fixture(scope="function")
def review_set(df_arguments):
    """List of arund 370 review texts."""
    reviews = df_arguments["argument"].tolist()
    return [r for r in reviews if r is not None]


@pytest.fixture(scope="function")
def topic_model():
    """TopicModel instance"""
    return TopicModel()


@pytest.fixture(scope="function")
def dummy_argument_selection():
    """Dummy argument dataframe."""
    return pd.DataFrame(
        {
            "argument": ["arg1", "arg2", "arg3", "arg4", "arg5"],
            "argument_id": [4, 7, 13, 29, 33],
            "score": [1, 2, 2, 3, 3],
            "coherence": [0.3, 0.5, 0.1, 0.05, 0.9],
        }
    )


@pytest.fixture(scope="function")
def dummy_edge_data():
    """Dummy edge dataframe."""
    return pd.DataFrame(
        {
            "source": [1, 0, 0, 4, 1, 4, 2, 4],
            "target": [0, 2, 3, 0, 3, 1, 3, 2],
            "weight": [0.2, 0.2, 0.25, 0.6, 0.45, 0.4, 0.05, 0.8],
        }
    )



"""Tests of the reader module"""
import pytest
import pandas as pd

from orangearg.argument.miner.reader import read_json_file


@pytest.fixture
def mocker_multi_objs(mocker):
    """Mock-open a file containing multiple JSON objects"""
    mocked_data = mocker.mock_open(
        read_data='{"col1": 1, "col2": 2}\n{"col1": 3, "col2": 4}\n'
    )
    mocker.patch("builtins.open", mocked_data)


@pytest.fixture
def mocker_semi_struct(mocker):
    """Mock-open a semi-structured JSON file"""
    mocked_data = mocker.mock_open(
        read_data='{"col1": 1, "col2": {"subcol1": 2, "subcol2": 3}}'
    )
    mocker.patch("builtins.open", mocked_data)


class TestReadJSONFile:
    """Tests of the read_json_file function"""

    def test_multi_object(self, mocker_multi_objs):
        """Test if the function can read a JSON file with multiple objects correctly."""
        data = read_json_file("fakefile")
        expected_data = pd.DataFrame({"col1": [1, 3], "col2": [2, 4]})
        assert data.equals(expected_data)

    def test_semi_struct(self, mocker_semi_struct):
        """Test if the function can read a semi-structure JSON file well"""
        data = read_json_file("fakefile")
        expected_data = pd.DataFrame(
            {"col1": [1], "col2.subcol1": [2], "col2.subcol2": [3]}
        )
        assert data.equals(expected_data)



"""Argument miner widget"""
from Orange.data import Table
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Input, Output, OWWidget
from Orange.data.pandas_compat import table_from_frame, table_to_frame
from AnyQt.QtCore import Qt

from ..miner.miner import (
    select_by_topic,
    get_edges,
    get_edge_weights,
    get_edge_table,
    get_node_labels,
    get_node_table,
)
from ..miner.utilities import check_columns


class OWArgMiner(OWWidget):
    """Argument miner widget

    Given a set of arguments (read from URL), mine the attacking relationship and generate
    the corresponding edge and node table.
    """

    name = "Argument Miner"
    description = (
        "Mine argument set and create edge and node tables of their attacking network."
    )
    icon = "icons/OWArgMiner.svg"
    want_main_area = False

    # GUI variables
    selected_topic = Setting(0)

    class Inputs:
        argument_data = Input("Argument Data", Table)
        topic_data = Input("Topic Data", Table)

    class Outputs:
        edge_data = Output("Edge Data", Table)
        node_data = Output("Node Data", Table)

    def __init__(self):
        super().__init__()

        # GUI
        box_select_topic = gui.widgetBox(
            self.controlArea, orientation=Qt.Vertical, box="Select Topic"
        )
        self.combo_topic = gui.comboBox(
            box_select_topic,
            self,
            "selected_topic",
            label="Select a topic",
            sendSelectedValue=False,
            items=(),
        )
        gui.button(
            widget=self.controlArea,
            master=self,
            label="Mine",
            callback=self.commit,
        )

    @Inputs.argument_data
    def set_argument_data(self, data):
        self.df_arguments = table_to_frame(data, include_metas=True)
        expected_cols = ["argument", "score", "topics", "sentiment", "coherence"]
        check_columns(expected_cols=expected_cols, data=self.df_arguments)

    @Inputs.topic_data
    def set_topic_data(self, data):
        self.df_topics = table_to_frame(data, include_metas=True)
        check_columns(expected_cols=["topic"], data=self.df_topics)

        topics = self.df_topics["topic"].sort_values(ascending=True)
        topics = topics.drop(labels=[0]).reset_index(drop=True)
        topics = topics.astype(str).tolist()

        # update combobox
        self.combo_topic.clear()
        self.combo_topic.addItems(topics)
        self.selected_topic = 0

    def commit(self):
        """Callback: commit button."""
        arg_selection = select_by_topic(
            data=self.df_arguments, topic=self.selected_topic
        )
        edges = get_edges(data=arg_selection)
        weights = get_edge_weights(data=arg_selection, edges=edges)
        df_edges = get_edge_table(edges=edges, weights=weights)

        labels = get_node_labels(
            indices=arg_selection.index.tolist(),
            sources=df_edges["source"].tolist(),
            targets=df_edges["target"].tolist(),
        )
        df_nodes = get_node_table(
            arg_ids=arg_selection["argument_id"].tolist(),
            arguments=arg_selection["argument"].tolist(),
            scores=arg_selection["score"].tolist(),
            labels=labels,
        )

        self.Outputs.edge_data.send(table_from_frame(df_edges))
        self.Outputs.node_data.send(table_from_frame(df_nodes))



"""Argument Explorer Widget"""
import numpy as np
import networkx as nx
from AnyQt.QtCore import Qt
from Orange.data import Table
from Orange.widgets import gui
from Orange.widgets.widget import Input
from Orange.widgets.visualize.utils.widget import OWDataProjectionWidget
from Orange.widgets.settings import SettingProvider, Setting
from Orange.widgets.utils.plot import OWPlotGUI
from Orange.data.pandas_compat import table_to_frame

from ..graph.graphview import GraphView

GRAPH_LAYOUT = ("spring", "multipartite", "kamada kawai", "spectral")


class OWArgExplorer(OWDataProjectionWidget):
    name = "Argument Explorer"
    description = "Visually explore arguments in their attacking network."
    icon = "icons/OWArgExplorer.svg"

    class Inputs:
        edge_data = Input("Edge Data", Table)
        node_data = Input("Node Data", Table)

    GRAPH_CLASS = GraphView
    graph = SettingProvider(GraphView)

    node_sparsity = Setting(5)
    graph_layout = Setting(
        GRAPH_LAYOUT[0]
    )  # comboBox widget returns index of the selection

    def __init__(self):
        super().__init__()

        self.edge_data = None
        self.node_data = None
        self.positions = None

    def _add_controls(self):
        self.gui = OWPlotGUI(self)
        layout = gui.vBox(self.controlArea, box="Layout")
        gui.comboBox(
            layout,
            self,
            "graph_layout",
            label="Graph layout",
            sendSelectedValue=True,
            items=GRAPH_LAYOUT,
            callback=self.relayout,
        )
        self.sparsity_control = gui.hSlider(
            layout,
            self,
            "node_sparsity",
            minValue=0,
            maxValue=10,
            intOnly=False,
            label="Node sparsity",
            orientation=Qt.Horizontal,
            callback_finished=self.relayout,
        )

    @Inputs.edge_data
    def set_edge_data(self, data):
        self.edge_data = data

    @Inputs.node_data
    def set_node_data(self, data):
        self.node_data = data

    def handleNewSignals(self):
        self.AlphaValue = 0
        self.relayout()

    def relayout(self):
        """recompute positions of nodes and reset the graph"""
        if self.node_data is None or self.edge_data is None:
            return

        self.sparsity_control.setEnabled(self.graph_layout == GRAPH_LAYOUT[0])
        self.set_positions()
        self.closeContext()
        self.data = self.node_data
        self.valid_data = np.full(len(self.data), True, dtype=bool)
        self.openContext(self.data)
        self.graph.reset_graph()

    def set_positions(self):
        """set coordinates of nodes to self.positions.

        Args:
            layout (str, optional): name of layout. Defaults to "sfdp".
        """
        df_edge = table_to_frame(self.edge_data)
        df_node = table_to_frame(self.node_data)

        G = nx.from_pandas_edgelist(
            df_edge,
            source="source",
            target="target",
            edge_attr=["weight"],
            create_using=nx.DiGraph(),
        )
        node_attrs = {i: {"subset": df_node["label"][i]} for i in G.nodes}
        nx.set_node_attributes(G, node_attrs)

        # in case arguments not appear in the attacking network
        if len(G.nodes) < df_node.shape[0]:
            remain_nodes = df_node.iloc[~df_node.index.isin(G.nodes)]
            G.add_nodes_from(remain_nodes.index.tolist())

        if self.graph_layout == GRAPH_LAYOUT[0]:
            spasity = (self.node_sparsity + 1) / 11.0
            pos_dict = nx.spring_layout(G, k=spasity, seed=10)
        elif self.graph_layout == GRAPH_LAYOUT[1]:
            pos_dict = nx.multipartite_layout(G)
        elif self.graph_layout == GRAPH_LAYOUT[2]:
            pos_dict = nx.kamada_kawai_layout(G)
        elif self.graph_layout == GRAPH_LAYOUT[3]:
            pos_dict = nx.spectral_layout(G)

        self.positions = []
        for i in sorted(pos_dict.keys()):
            self.positions.append(pos_dict[i])
        self.positions = np.array([*self.positions])

    def get_embedding(self):
        return self.positions

    def get_edges(self):
        return table_to_frame(self.edge_data)

    def get_marked_nodes(self):
        return None

    def get_node_labels(self):
        return table_to_frame(self.node_data)["label"]

    def selection_changed(self):
        super().selection_changed()
        self.graph.update_edges()



"""JSON Reader widget."""
import os

from Orange.data import Table
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Output, OWWidget
from Orange.data.pandas_compat import table_from_frame
from AnyQt.QtWidgets import QFileDialog
from AnyQt.QtCore import Qt

from ..miner.reader import read_json_file


class OWJSONReader(OWWidget):
    """Json file reader widget

    This widget reads a local JSON file and output the data as table.
    """

    name = "JSON Reader"
    description = "Read a local JSON file and output the data as table."
    icon = "icons/OWJSONReader.svg"
    want_main_area = False

    # GUI variables
    fpath = Setting("")

    class Outputs:
        output_data = Output("Data", Table)

    def __init__(self):
        super().__init__()

        # GUI
        box_select_file = gui.widgetBox(
            self.controlArea, orientation=Qt.Horizontal, box="Select File"
        )
        gui.button(box_select_file, self, "...", callback=self.browse_file)
        self.label_path = gui.label(
            box_select_file, self, "Path: %(fpath)s", labelWidth=300
        )
        gui.button(self.controlArea, self, "Read", callback=self.read)

    def browse_file(self):
        """Call back: enable file dialog and get the path of the select file."""
        start_dir = os.path.expanduser("./")
        self.fpath, _ = QFileDialog.getOpenFileName(self, "Open File", start_dir, "")

        # format the path label
        text = self.label_path.text()
        text_len_limit = 40
        if len(text) > text_len_limit:
            text = "..." + text[-text_len_limit:]
            self.label_path.setText(text)

    def read(self):
        """Call back: read file from json and output dataframe"""
        data = read_json_file(self.fpath)
        data = data.dropna().reset_index(drop=True)
        table = table_from_frame(data)
        self.Outputs.output_data.send(table)



"""Argument mining module"""
from ast import literal_eval
from typing import Tuple, List
from itertools import combinations

import pandas as pd

from orangearg.argument.miner.utilities import check_columns


def select_by_topic(data: pd.DataFrame, topic: int) -> pd.DataFrame:
    """Select arguments mentioning the given topic.

    Args:
        data (pd.DataFrame): The argument dataframe that must contain the 'topics' column.
        topic (int): The given topic to select.

    Raises:
        ValueError: if the 'topics' value of an argument is stored as something else other than a tuple (e.g. a list).

    Returns:
        pd.DataFrame: Part of the original argument dataframe that only contains arguments mentioning the given topic.
    """
    expected_cols = ["topics"]
    check_columns(expected_cols=expected_cols, data=data)

    select_condition = []
    for i, row in data.iterrows():
        topics = literal_eval(str(row["topics"]))
        if isinstance(topics, int):  # in case of tuple of one item
            select_condition.append(topic == topics)
        elif isinstance(topics, tuple):
            select_condition.append(topic in topics)
        else:
            raise ValueError(
                f"Topics of the {i}th argument should be a tuple, but {topics} is given."
            )

    selection = data[select_condition]
    selection["argument_id"] = selection.index
    return selection.reset_index(drop=True)


def get_edges(data: pd.DataFrame) -> List[Tuple[int]]:
    """Get edges from argument dataframe.

    Edges (attacks) only exist if the two arguments have different overall scores. Edges are tuple of source and target, which are indices of the corresponding argument in the input dataframe.

    Args:
        data (pd.DataFrame): The argument dataframe that must have the 'score' column.

    Returns:
        List[Tuple[int]]: The edge list.
    """
    expected_cols = ["score"]
    check_columns(expected_cols=expected_cols, data=data)

    id_combs = list(combinations(data.index, 2))
    edges = []
    for id_combo in id_combs:
        if data.loc[id_combo[0]]["score"] != data.loc[id_combo[1]]["score"]:
            edges.append(id_combo)
    return edges


def get_edge_weights(data: pd.DataFrame, edges: List[Tuple[int]]) -> List[float]:
    """Get edge weights.

    Edge weights are computed as the difference between the coherence of the source and that of the target.

    Args:
        data (pd.DataFrame): The argument dataframe that must have the 'coherence' column.
        edges (List[Tuple[int]]): The edge list.

    Returns:
        List[float]: The list of edge weights.
    """
    expected_cols = ["coherence"]
    check_columns(expected_cols=expected_cols, data=data)

    weights = []
    for s, t in edges:
        weight = data.loc[s]["coherence"] - data.loc[t]["coherence"]
        weights.append(round(weight, 2))
    return weights


def get_edge_table(edges: List[Tuple[int]], weights: List[float]) -> pd.DataFrame:
    """Get the edge dataframe.

    There will be three columns in the output dataframe, which are 'source', 'target', and 'weight'. Together, they describe weighted directed edges from source to target argument. Note that there will be no negative weights in the output dataframe, instead, all values will be replace with their absolute values. For edges with negative weights, we swap their source and target.

    Args:
        edges (List[Tuple[int]]): The edge list, which are tuples of source and target argument ids.
        weights (List[float]): The list of edge weights.

    Raises:
        ValueError: if size of the input lists doesn't match.

    Returns:
        pd.DataFrame: The result edge dataframe.
    """
    if len(edges) != len(weights):
        raise ValueError(
            f"Length of edges and weigts are not equal: {(len(edges), len(weights))}."
        )

    for i, w in enumerate(weights):
        if w < 0:
            edges[i] = (edges[i][1], edges[i][0])
    result = pd.DataFrame(edges, columns=["source", "target"])
    result["weight"] = [abs(w) for w in weights]
    return result


def get_node_labels(
    indices: List[int], sources: List[int], targets: List[int]
) -> List[str]:
    """Get labels of arguments given the attacking network.

    Arguments are separated into two classes, 'supportive' and 'defeated', which generally means reliable and unreliable. The rule of detecting the labels is as follows: if an argument is attacked by another argument who is not attacked by any argument, then this argument is labeled as 'defeated'; otherwise, it's labeled as 'supportive'. That means, if an argument appears in `targets`, where its corresponding source doesn't, this argument will be labeled as 'defeated', and otherwise 'supportive'.

    Args:
        indices (List[int]): The node index list
        sources (List[int]): The source list of the attacking network.
        targets (List[int]): The target list of the attacking network.

    Returns:
        List[str]: The label list.
    """
    labels = {i: "supportive" for i in indices}

    for i, target in enumerate(targets):
        source = sources[i]
        if source not in targets:
            labels[target] = "defeated"

    return list(labels.values())


def get_node_table(
    arg_ids: List[int], arguments: List[str], scores: List[int], labels: List[str]
) -> pd.DataFrame:
    """Get the node dataframe.

    The node dataframe will contain 4 columns, that are 'argument_id', 'argument', 'score', and 'label'.

    Args:
        arg_ids (List[int]): The argument id list.
        arguments (List[str]): The argument text list.
        scores (List[int]): The list of argument overall score.
        labels (List[str]): The argument label list.

    Returns:
        pd.DataFrame: The result node dataframe.
    """
    return pd.DataFrame(
        {
            "argument_id": arg_ids,
            "argument": arguments,
            "score": scores,
            "label": labels,
        }
    )



"""Test the processor module"""
from contextlib import nullcontext as does_not_raise

import pandas as pd
import pytest
from pytest import approx

from orangearg.argument.miner.processor import (
    get_argument_topics,
    get_argument_sentiment,
    get_argument_coherence,
    update_argument_table,
    _match_list_size,
    _aggregate_list_by_another,
)
from orangearg.argument.miner.utilities import check_columns


@pytest.mark.parametrize(
    "dummy_list1, dummy_list2, exception_context",
    [
        ([1], [1], does_not_raise(enter_result=None)),
        ([1], [1, 2], pytest.raises(ValueError)),
    ],
)
def test__match_list_size(dummy_list1, dummy_list2, exception_context):
    """Unit test _match_list_size"""
    with exception_context as ex:
        _match_list_size(dummy_list1, dummy_list2)
    if ex:
        assert str(ex.value) == f"Input size not match: {(dummy_list1, dummy_list2)}."


def test__aggregate_list_by_another():
    """Unit test _aggregate_list_by_another."""
    dummy_keys = [1, 1, 2, 2]
    dummy_values = ["A", "B", "C", "D"]

    result = _aggregate_list_by_another(keys=dummy_keys, values=dummy_values)
    assert result == {1: ["A", "B"], 2: ["C", "D"]}


def test_get_argument_topics():
    """Unit test get_argument_topics."""
    dummy_ids = [0, 0, 1, 1, 1]
    dummy_topics = [-1, 0, 0, 1, 2]
    topics = get_argument_topics(arg_ids=dummy_ids, topics=dummy_topics)

    assert topics == [(-1, 0), (0, 1, 2)]


def test_get_argument_sentiment():
    """Unit test get_argument_sentiment."""
    dummy_ids = [0, 0, 1, 1, 1]
    dummy_ranks = [0.3, 0.7, 0.25, 0.2, 0.55]
    dummy_p_scores = [-0.5, 0.2, 0.3, 0.7, -0.45]
    sentiments = get_argument_sentiment(
        arg_ids=dummy_ids, ranks=dummy_ranks, p_scores=dummy_p_scores
    )

    assert sentiments == approx([0.495, 0.484], 0.01)


def test_get_argument_coherence():
    """Unit test get_argument_coherence"""
    dummy_sentiments = [-0.5, 0.8, 1, 0]
    dummy_scores = [1, 3, 2, 5]
    coherences = get_argument_coherence(
        sentiments=dummy_sentiments, scores=dummy_scores
    )

    assert coherences == approx([0.535, 0.799, 0.245, 0.082], 0.01)


def test_update_argument_table(mocker):
    """Unit test update_argument_table"""
    dummy_topics = [[1, 2], [3, 4]]
    dummy_sentiments = [0.4, 0.8]
    dummy_coherences = [0.1, 0.3]
    dummy_df_arguments = pd.DataFrame()
    expected_result = pd.DataFrame(
        {
            "topics": dummy_topics,
            "sentiment": dummy_sentiments,
            "coherence": dummy_coherences,
        }
    )
    mock_deepcopy = mocker.patch("copy.deepcopy", return_value=dummy_df_arguments)

    result = update_argument_table(
        df_arguments=dummy_df_arguments,
        topics=dummy_topics,
        sentiments=dummy_sentiments,
        coherences=dummy_coherences,
    )

    assert result.equals(expected_result)
    mock_deepcopy.assert_called_once_with(dummy_df_arguments)


def test_integrate_processor(df_arguments, df_chunks):
    """Integrate test the processor module."""
    expected_cols_chunks = ["argument_id", "topic", "rank", "polarity_score"]
    expected_cols_arguments = ["score"]

    check_columns(expected_cols=expected_cols_chunks, data=df_chunks)
    check_columns(expected_cols=expected_cols_arguments, data=df_arguments)

    chunk_arg_ids = df_chunks["argument_id"]
    chunk_topics = df_chunks["topic"]
    chunk_ranks = df_chunks["rank"]
    chunk_p_scores = df_chunks["polarity_score"]
    arg_scores = df_arguments["score"]

    arg_topics = get_argument_topics(arg_ids=chunk_arg_ids, topics=chunk_topics)
    arg_sentiments = get_argument_sentiment(
        arg_ids=chunk_arg_ids, ranks=chunk_ranks, p_scores=chunk_p_scores
    )
    arg_coherences = get_argument_coherence(
        scores=arg_scores, sentiments=arg_sentiments
    )
    assert all([0.0 <= s <= 1.0 for s in arg_sentiments])
    assert all([0.0 <= c <= 1.0 for c in arg_coherences])
    assert (
        len(arg_topics)
        == len(arg_sentiments)
        == len(arg_coherences)
        == df_arguments.shape[0]
    )

    df_arguments_updated = update_argument_table(
        df_arguments=df_arguments,
        topics=arg_topics,
        sentiments=arg_sentiments,
        coherences=arg_coherences,
    )



"""Argument filre reader module

This module implements functions for reading input data files in
different formats. So far, we only have the support to JSON file.
But we forsee the need of supporting other formats, and all future
functions in this scope should be in this module.
"""

import json

import pandas as pd


def read_json_file(fpath: str) -> pd.DataFrame:
    """Read a local JSON file and return its content as a pandas dataframe.

    This function will automatically handle the case that a JSON
    file contains multiple JSON objects. It will also normalize
    semi-structured JSON strings.

    Args:
        fpath (str): The file path

    Returns:
        pd.DataFrame: The pandas dataframe object that contains
        content of the JSON file read from the given path.

    """

    with open(fpath, "r", encoding="utf-8") as json_file:
        json_data = []
        # handle multiple JSON objects in one file
        for json_object in json_file:
            json_data.append(json.loads(json_object))

    # normalize semi-structured json as flat table
    return pd.json_normalize(json_data)



"""Argument processor module."""

from typing import List, Dict, Tuple
import copy
import math
from collections import defaultdict

import numpy as np
import pandas as pd


def _match_list_size(*args: List):
    """With an arbitrary number of lists as input, check if they are in the same size."""
    if any(len(arg) != len(args[0]) for arg in args):
        raise ValueError(f"Input size not match: {args}.")


def _aggregate_list_by_another(keys: List, values: List) -> Dict:
    """Aggregate a list according to elements of another list.

    Args:
        keys (List): The group keys.
        values (List): The list to be aggregated.

    Returns:
        Dict: The aggregation result.
    """
    result = defaultdict(list)
    for i, key in enumerate(keys):
        result[key].append(values[i])
    return result


def get_argument_topics(arg_ids: List[int], topics: List[int]) -> List[Tuple[int]]:
    """Get argument topics.

    The topics of an argument is a combination of the topics of all chunks that belong to this argument. Duplications are not removed, and the reason behind is that duplications can be treated as a sign of topic importance. Also, even though two chunks can belong to the same topic, they could still have different ranks within an argument.

    Args:
        arg_ids (List[int]): the argument ids of chunks.
        topics (List[int]): the topic indices of chunks.

    Returns:
        List[list[int]]: list of argument topics, which is also a list containing topic indices of chunks belonging to this argument.
    """
    _match_list_size(arg_ids, topics)
    result = _aggregate_list_by_another(keys=arg_ids, values=topics)
    result = result.values()
    return [tuple(r) for r in result]


def get_argument_sentiment(
    arg_ids: List[int],
    ranks: List[float],
    p_scores: List[float],
    min_sent: int = -1,
    max_sent: int = 1,
) -> List[float]:
    """Get argument sentiment score.

    The sentiment score of an argument is calculated as a weighted sum of sentiment scores of chunks belonging to this argument, where weights are ranks of the chunks. The result score is then normalized into range [0, 1].

    Args:
        arg_ids (List[int]): the argument ids of chunks.
        ranks (List[float]): the pagerank of chunks within arguments.
        p_scores (List[float]): the sentiment polarity scores of chunks.
        min_sent (int): minimun of argument sentiment before normalization. Defaults to -1.
        max_sent (int): maximum of argument sentiment before normalization. Defaults to 1.

    Returns:
        List[float]: List of argument sentiment scores, which are floats in range [0, 1].
    """
    _match_list_size(arg_ids, ranks, p_scores)

    grouped_ranks = _aggregate_list_by_another(keys=arg_ids, values=ranks)
    grouped_p_scores = _aggregate_list_by_another(keys=arg_ids, values=p_scores)

    sentiments = []
    for arg_id, rank in grouped_ranks.items():
        p_score = grouped_p_scores[arg_id]
        sentiment = np.dot(rank, p_score)
        sentiment = (sentiment - min_sent) / (max_sent - min_sent)
        sentiments.append(sentiment)
    return sentiments


def get_argument_coherence(
    scores: List[int],
    sentiments: List[float],
    min_score: int = 1,
    max_score: int = 5,
    variance: float = 0.2,
) -> List[float]:
    """Get argument coherence.

    Coherence is computed as inversed difference between sentiments and overall scores. Overall scores are first normalized into the same range as argument sentiments, which is [0, 1]. Then their differences are computed and applied a Gaussian kernal to invert and scale the differences to [0, 1].

    Args:
        scores (List[int]): List of argument overall scores.
        sentiments (List[float]): List of argument sentiment scores.
        min_score (int, optional): Lower bound of scores. Defaults to 1.
        max_score (int, optional): Upper bound of scores. Defaults to 5.
        variance (float): variance of the Gaussian kernal.

    Returns:
        List[float]: List of argument coherence scores, in range of (0, 1]
    """
    _match_list_size(sentiments, scores)

    range_score = max_score - min_score
    scores = [(s - min_score) / range_score for s in scores]

    def gaussian(x):
        """Gaussian activation function."""
        return math.e ** (-(x**2) / (2 * variance))

    coherences = [sentiments[i] - scores[i] for i in range(len(scores))]
    coherences = list(map(gaussian, coherences))
    return coherences


def update_argument_table(
    df_arguments: pd.DataFrame,
    topics: List[List[int]],
    sentiments: List[float],
    coherences: List[float],
) -> pd.DataFrame:
    """Return a copy of argument dataframe, with new columns of argument topics, sentiments, and coherences.

    Args:
        df_arguments (pd.DataFrame): argument dataframe.
        topics (List[List[int]]): list of argument topics
        sentiments (List[float]): list of argument sentiment scores
        coherences (List[float]): list of argument coherence scores

    Returns:
        pd.DataFrame: _description_
    """
    df_copy = copy.deepcopy(df_arguments)
    df_copy["topics"] = topics
    df_copy["sentiment"] = sentiments
    df_copy["coherence"] = coherences
    return df_copy



"""Argument chunker module"""

from typing import List, Tuple
import itertools
import logging

import pandas as pd
import numpy as np
from sentence_transformers import SentenceTransformer
from umap import UMAP
from hdbscan import HDBSCAN
from sklearn.feature_extraction.text import CountVectorizer
from bertopic import BERTopic
from bertopic.vectorizers import ClassTfidfTransformer
from bertopic.representation import PartOfSpeech
import spacy
import networkx as nx
from textblob import TextBlob


def load_nlp_pipe(model_name: str):
    """Download the required nlp pipe if not exist

    Args:
        model_name (str): name of the nlp pipe, a full list of models can be found from https://spacy.io/usage/models.

    Returns:
        The spacy nlp model.
    """
    try:
        spacy.load(name=model_name)
    except OSError:
        spacy.cli.download(model=model_name)
    return spacy.load(name=model_name)


def get_chunk(docs: List[str]) -> Tuple[List[int], List[str]]:
    """Split documents of a given corpus into chunks.

    A chunk can be considered as a meaningful clause, which can be part of a sentence. For instance, the sentence "I like the color of this car but it's too expensive." will be splitted as two chunks, which are "I like the color of this car" and "but it's too expensive". A dependency parser is implemented for doing this job.

    Args:
        docs (List[str]): The input corpus.

    Returns:
        List[int]: ids of the arguments that the chunks belongs to.
        List[str]: chunk text.
    """
    arg_ids, chunks = [], []

    def create_chunk(words, arg_id):
        """Merge words into chunk and add to the output lists"""
        chunk = " ".join([w.text for w in words])
        arg_ids.append(arg_id)
        chunks.append(chunk)

    def find_heads(sentence):
        """Identify the sentence heads. Currently, the strategy is quite basic: it involves designating as heads all words whose dependencies are conjunctions. However, there is room for enhancement in the future."""
        heads = []
        for word in sentence.root.children:
            if word.dep_ == "conj":
                heads.append(word)
        return heads

    nlp = load_nlp_pipe(model_name="en_core_web_md")
    for i, doc in enumerate(docs):
        if doc is not None:
            doc = nlp(doc)
            for sentence in doc.sents:
                seen = set()
                heads: list[spacy.tokens.span] = find_heads(sentence)
                for head in heads:
                    head_phrase = list(head.subtree)
                    seen.update(head_phrase)
                    create_chunk(words=head_phrase, arg_id=i)
                unseen = [w for w in sentence if w not in seen]
                create_chunk(words=unseen, arg_id=i)
    return arg_ids, chunks


def get_chunk_polarity_score(chunks: List[str]):
    """Compute polarity score of each chunk in the given list.

    The polarity score is a float within the range [-1.0, 1.0], where 0 means neutral, + means positive, and - means negative.

    Args:
        chunks (List[str]): chunk list

    Returns:
        List[float]: polarity scores of the given chunks
    """
    p_scores = []
    for chunk in chunks:
        p_scores.append(TextBlob(chunk).sentiment.polarity)
    return p_scores


def get_chunk_topic(chunks: List[str]):
    """Get topic information and embedding vectors of chunks via topic modeling.

    Args:
        chunks (List[str]): chunk list.

    Returns:
        List[int]: topic ids of chunks.
        np.ndarray: embedding vectors of chunks.
        pd.DataFrame: Table of topic information.
    """
    topic_model = TopicModel()
    topics = topic_model.fit_transform_reduced(chunks)
    embeds = topic_model.get_doc_embeds()
    df_topics = topic_model.get_topic_table()
    return topics, embeds, df_topics


def get_chunk_rank(arg_ids: List[int], embeds: np.ndarray):
    """In each argument, comput rank of chunks within.

    Rank can be understood as importance of chunks. This function computes the relative importance of chunks within arguments they belong to. This is done by applying the Pagerank algorithm, where similarity is computed as the cosine similarity of chunk embedding vectors.

    Args:
        arg_ids (List[int]): ids of arguments that chunks belongs to.
        embeds (np.ndarray): embedding vectors of chunks.

    Returns:
        List[float]: rank of chunks
    """

    def rank_in_argument(embeds: pd.Series):
        """Compute pagerank of embedding vectors."""
        embeds = np.array(list(embeds))
        embeds /= np.expand_dims(np.linalg.norm(embeds, axis=-1), axis=-1)
        sim_mat = embeds @ embeds.T
        graph = nx.from_numpy_array(sim_mat)
        ranks = list(nx.pagerank(graph).values())
        ranks = np.array(ranks)
        return list(ranks)

    df_embeds = pd.DataFrame({"arg_id": arg_ids, "embed": embeds.tolist()})
    df_embeds = df_embeds.groupby(by="arg_id", as_index=False).agg(
        {"embed": rank_in_argument}
    )  # group by arg_id to make the ranking within arguments
    ranks = df_embeds["embed"].tolist()
    ranks = list(itertools.chain(*ranks))
    return ranks


def get_chunk_table(
    arg_ids: List[int],
    chunks: List[str],
    p_scores: List[float],
    topics: List[int],
    ranks: List[float],
):
    """Given all the measures of chunks, generate and return the chunk table as a pandas dataframe, with pre-defined column names.

    Args:
        arg_ids (List[int]): ids of arguments that chunks belong to
        chunks (List[str]): chunk text
        p_scores (List[float]): polarity score of chunks
        topics (List[int]): topic id of chunks
        ranks (List[float]): rank of chunks

    Returns:
        pd.DataFrame: chunk table
    """
    return pd.DataFrame(
        {
            "argument_id": arg_ids,
            "chunk": chunks,
            "polarity_score": p_scores,
            "topic": topics,
            "rank": ranks,
        }
    )


class TopicModel:
    """Topic modeling class.

    Functions are implemented based on the BERTopic model. For now, the topic model is setup with a set of default parameters of the sub-models. However, it should be possible that the user can config it further. This will be a next step.

    Attributes:
        _rd_model (:obj:'UMAP'): instance of UMAP algorithm as the dimensionality reduction sub-model.
        model (:obj:'BERTopic'): the topic model that applied the sub-models predefined.
    """

    def __init__(self):
        self._rd_model = None
        self.model = None

        # initialize the topic model
        self.init_model()

    def init_model(
        self,
        transformer: str = "all-mpnet-base-v1",
        n_components: int = 5,
        min_cluster_size: int = 10,
        ngram_min: int = 1,
        ngram_max: int = 1,
    ):
        """Initialize the topic model by indicating a number of arguments.

        Args:
            transformer (str, optional): Name of the sentence embedding model. Defaults to "all-mpnet-base-v1". A list of pretrained models can be found here: https://www.sbert.net/docs/pretrained_models.html.
            n_components (int, optional): Number of dimensions after reduction. Defaults to 5.
            min_cluster_size (int, optional): Minimum size of clusters for the clustering algorithm. Defaults to 5.
            ngram_min (int, optional): Low band of ngram range for topic representation. Defaults to 1.
            ngram_max (int, optional): High band of ngram range for topic representation. Defaults to 1.
        """
        language = "english"
        nlp_pipe = "en_core_web_md"
        load_nlp_pipe(model_name=nlp_pipe)

        # initialize sub-models
        embed_model = SentenceTransformer(model_name_or_path=transformer)
        self._rd_model = UMAP(n_components=n_components)
        cluster_model = HDBSCAN(min_cluster_size=min_cluster_size, prediction_data=True)
        vector_model = CountVectorizer(
            stop_words=language, ngram_range=[ngram_min, ngram_max]
        )
        ctfidf_model = ClassTfidfTransformer(
            reduce_frequent_words=True,
        )
        represent_model = PartOfSpeech(
            model=nlp_pipe,
            pos_patterns=[[{"POS": "NOUN"}], [{"POS": "ADJ"}], [{"POS": "VERB"}]],
        )

        # initialize topic model
        self.model = BERTopic(
            language=language,
            embedding_model=embed_model,
            umap_model=self._rd_model,
            hdbscan_model=cluster_model,
            vectorizer_model=vector_model,
            ctfidf_model=ctfidf_model,
            representation_model=represent_model,
            calculate_probabilities=False,
        )

    def fit_transform_reduced(self, docs: List[str]) -> List[int]:
        """Further reduce outliers from the result of the fit_transform function.

        Note that BERTopic is a clustering approach, which means that it doesn not work if there is nothing to be clustered. And keep in mind that the input corpus should contain at least 1000 documents to get meaningful results. Refer to this thread: https://github.com/MaartenGr/BERTopic/issues/59#issuecomment-775718747.

        Args:
            docs (List[str]): The input corpus.

        Returns:
            List[int]: Topics of the input docs.
        """
        if len(docs) < 1000:
            logging.warning(
                "The input corpus should contain at least 1000 documents to get meaningful results. Got %d.",
                len(docs),
            )
        topics, _ = self.model.fit_transform(docs)
        try:
            new_topics = self.model.reduce_outliers(docs, topics, strategy="embeddings")
        except ValueError as e:
            logging.error("Failed to reduce outliers: %s", str(e))
            new_topics = topics

        self.model.update_topics(docs, topics=new_topics)
        return new_topics

    def get_topic_table(self) -> pd.DataFrame:
        """Get the table of topic information and return it as a pandas dataframe.

        Returns:
            pd.DataFrame: The topic table.
        """
        topic_info = self.model.get_topic_info()
        topic_info = topic_info.drop(["Representative_Docs"], axis=1)
        topic_info = topic_info.rename(
            columns={
                "Topic": "topic",
                "Count": "count",
                "Name": "name",
                "Representation": "keywords",
            }
        )
        topic_info["keywords"] = topic_info["keywords"].apply(tuple)
        return topic_info

    def get_doc_embeds(self) -> np.ndarray:
        """Get the embeddings of the docs.

        Returns:
            np.ndarray: Embeddings of the docs, in size of (n_doc, n_components).
        """
        return self._rd_model.embedding_



"""Tests of the chunker module"""
from difflib import SequenceMatcher

import pytest
import numpy as np
import pandas as pd

from orangearg.argument.miner.chunker import (
    load_nlp_pipe,
    get_chunk,
    get_chunk_polarity_score,
    get_chunk_rank,
    get_chunk_topic,
    get_chunk_table,
)
from .conftest import large_chunk_set, review_set, topic_model


def test_load_nlp_pipe(mocker):
    """Unit test load_nlp_pipe."""

    def check_download(name):
        """When executing effect as a callable, it will need to pass in all the args and kwargs of the mocked function, thus we have a `name` kwarg as a placeholder, otherwise won't work."""
        if not mock_spacy_cli_download.called:
            raise (OSError)
        else:
            return "dummy_model"

    mock_spacy_load = mocker.patch(
        "spacy.load",
        side_effect=check_download,
    )
    mock_spacy_cli_download = mocker.patch("spacy.cli.download")

    nlp = load_nlp_pipe(model_name="dummy_name")

    assert nlp == "dummy_model"
    mock_spacy_load.assert_any_call(name="dummy_name")
    mock_spacy_cli_download.assert_called_once_with(model="dummy_name")


class TestGetChunk:
    """Tests for testing the get_chunk function."""

    test_data = [
        # cases of coordinating conjunction
        (
            [
                "This is my first car and I'm satisfied with it.",
                "I have two goldfish and a cat.",
            ],
            [0, 0, 1],
            [
                "I'm satisfied with it",
                "This is my first car and",
                "I have two goldfish and a cat",
            ],
        ),
        # cases of correlative conjunction
        (
            [
                "I will either go for a hike or stay home and watch TV.",
                "The kid was running as fast as us.",
            ],
            [0, 0, 1],
            [
                "stay home and watch TV",
                "I will either go for a hike or",
                "The kid was running as fast as us",
            ],
        ),
        # cases of subordinating conjunction
        (
            [
                "Unless we give him a ride, he will not be able to come.",
                "I'm not going to work because I am sick.",
            ],
            [0, 1],
            [
                "Unless we give him a ride he will not be able to come",
                "I'm not going to work because I am sick",
            ],
        ),
        # case of multiple sentences in one document
        (
            [
                "The suspension system is very poor. Driving on slightly uneven roads feels very bumpy."
            ],
            [0, 0],
            [
                "The suspension system is very poor",
                "Driving on slightly uneven roads feels very bumpy",
            ],
        ),
    ]

    def similarity(self, a: str, b: str) -> float:
        """Match two strings and return their similarity."""
        a = a.lower()
        b = b.lower()
        match = SequenceMatcher(a=a, b=b)
        return match.ratio()

    @pytest.mark.parametrize("docs, expected_arg_ids, expected_chunks", test_data)
    def test_get_chunks(self, docs, expected_arg_ids, expected_chunks):
        """Test of dealing with cases containing coordinating conjunctions. The result chunks might look a bit different from the expected chunks, since marks and the conjunctions themselves are not considered. Therefore, a similarity check is given to make sure that the most part fo the result chunks will meet the expectation."""
        arg_ids, chunks = get_chunk(docs=docs)

        assert arg_ids == expected_arg_ids
        for i, _ in enumerate(chunks):
            assert self.similarity(chunks[i], expected_chunks[i]) >= 0.9


def test_get_chunk_polarity_score():
    """Unit test get_chunk_polarity_score."""
    chunks = [
        "The shirt looks great!",  # extremely positive
        "The shirt looks awful!",  # extremely negative
        "Although I don't like it that much, I will still buy it since it's so cheap.",  # relatively positive
        "Although I like it so much, I won't buy it since it's too expensive.",  # relatively negative
        "That car is made in Japan.",  # neutral
    ]
    scores = get_chunk_polarity_score(chunks=chunks)

    assert scores[0] > 0.5
    assert scores[1] < -0.5
    assert 0 < scores[2] < 0.5
    assert -0.5 < scores[3] < 0
    assert scores[4] == 0


def test_get_chunk_rank():
    """Unit test get_chunk_rank."""
    arg_ids = [0, 0, 0, 0]
    embeds = np.array(
        [
            [0, 1],
            [0.2, 0.8],
            [0.4, 1.6],
            [1, 0],
        ]
    )
    ranks = get_chunk_rank(arg_ids=arg_ids, embeds=embeds)

    assert len(ranks) == 4
    assert ranks[1] == ranks[2] > ranks[0] > ranks[3]


def test_get_chunk_topic(mocker):
    """Unit test get_chunk_topic."""
    dummy_chunks = ["chunk1", "chunk2", "chunk3"]
    loc_str = "orangearg.argument.miner.chunker.TopicModel."

    mock_fit_transform_reduced = mocker.patch(
        loc_str + "fit_transform_reduced", return_value="dummy_topics"
    )
    mock_get_topic_table = mocker.patch(
        loc_str + "get_topic_table", return_value="dummy_table"
    )
    mock_get_doc_embeds = mocker.patch(
        loc_str + "get_doc_embeds", return_value="dummy_embeds"
    )

    topics, embeds, df_topics = get_chunk_topic(dummy_chunks)

    assert topics == "dummy_topics"
    assert embeds == "dummy_embeds"
    assert df_topics == "dummy_table"
    mock_fit_transform_reduced.assert_called_once_with(dummy_chunks)
    mock_get_topic_table.assert_called_once()
    mock_get_doc_embeds.assert_called_once()


def test_get_chunk_table():
    """Unit test get_chunk_table"""
    arg_ids = [0, 0, 1, 1]
    chunks = ["chunk0", "chunk1", "chunk2", "chunk3"]
    p_scores = [0.1, -0.5, 0, 0.3]
    topics = [0, 1, 1, 0]
    ranks = [0.2, 0.8, 0.5, 0.5]
    df_result = get_chunk_table(
        arg_ids=arg_ids, chunks=chunks, p_scores=p_scores, topics=topics, ranks=ranks
    )
    expected_cols = ["argument_id", "chunk", "polarity_score", "topic", "rank"]

    assert df_result.shape == (4, 5)
    assert sorted(df_result.columns) == sorted(expected_cols)


def test_integrate_chunker(review_set):
    """Integration test functions of the chunker module."""
    arg_ids, chunks = get_chunk(docs=review_set)
    p_scores = get_chunk_polarity_score(chunks=chunks)
    topics, embeds, df_topics = get_chunk_topic(chunks=chunks)
    ranks = get_chunk_rank(arg_ids=arg_ids, embeds=embeds)
    df_chunks = get_chunk_table(
        arg_ids=arg_ids,
        chunks=chunks,
        p_scores=p_scores,
        topics=topics,
        ranks=ranks,
    )

    assert len(set(arg_ids)) == len(review_set), [
        i for i in range(len(review_set)) if i not in arg_ids
    ]
    assert len(chunks) == len(p_scores) == len(topics) == len(embeds) == len(ranks)
    assert len(set(topics)) == df_topics.shape[0]
    assert all(-1.0 <= p <= 1.0 for p in p_scores)
    assert all(0.0 <= r <= 1.0 for r in ranks)
    assert isinstance(df_topics, pd.DataFrame)
    assert isinstance(df_chunks, pd.DataFrame)


class TestTopicModel:
    """Tests of the TopicModel class."""

    def test_init_model(self, topic_model):
        """Unit test init_model."""
        topic_model.init_model()
        assert topic_model.model is not None

    def test_fit_transform_reduced(self, mocker, topic_model):
        """Unit test fit_transform_reduced."""
        size = 10
        chunks = ["This is dummy chunk%d." % i for i in range(size)]
        expected_topics = [-1] * size
        mock_fit_transform = mocker.patch.object(
            topic_model.model,
            "fit_transform",
            return_value=(expected_topics, None),
        )
        mock_reduce_outliers = mocker.patch.object(
            topic_model.model,
            "reduce_outliers",
            side_effect=ValueError,
        )
        mock_update_topics = mocker.patch.object(topic_model.model, "update_topics")

        result = topic_model.fit_transform_reduced(chunks)

        assert result == expected_topics
        mock_fit_transform.assert_called_once_with(chunks)
        mock_reduce_outliers.assert_called_once_with(
            chunks, expected_topics, strategy="embeddings"
        )
        mock_update_topics.assert_called_once_with(chunks, topics=expected_topics)

    def test_get_topic_table(self, mocker, topic_model):
        """Unit test get_topic_table."""
        mock_get_topic_info = mocker.patch.object(
            topic_model.model,
            "get_topic_info",
            return_value=pd.DataFrame(
                {
                    "Representative_Docs": ["doc1"],
                    "Representation": [["keyword1", "keyword2"]],
                }
            ),
        )
        mock_rename = mocker.patch(
            "pandas.DataFrame.rename",
            return_value=pd.DataFrame({"keywords": [["keyword1", "keyword2"]]}),
        )

        result = topic_model.get_topic_table()

        assert isinstance(result, pd.DataFrame)
        assert result.compare(
            pd.DataFrame({"keywords": [("keyword1", "keyword2")]})
        ).empty
        mock_get_topic_info.assert_called_once()
        mock_rename.assert_called_once()

    def test_get_doc_embeds(self, mocker, topic_model):
        """Unit test get_doc_embeds."""
        topic_model._rd_model = mocker.Mock(embedding_=np.array([]))

        result = topic_model.get_doc_embeds()

        assert isinstance(result, np.ndarray)

    def test_integrate_topic_model(self, large_chunk_set, topic_model):
        """Integration test of the Topic Model class."""
        topics = topic_model.fit_transform_reduced(large_chunk_set)
        embeds = topic_model.get_doc_embeds()
        df_topics = topic_model.get_topic_table()

        assert len(large_chunk_set) == len(topics) == len(embeds)
        assert len(embeds[0]) == 5  # if size of embeddings aligns to n_component
        assert isinstance(df_topics, pd.DataFrame)
        assert df_topics.shape[0] > 2  # at least 2 clusters
        expected_cols = ["topic", "count", "name", "keywords"]
        assert sorted(df_topics.columns) == sorted(expected_cols)



"""Collection of helper functions."""
from typing import List

import pandas as pd


def check_columns(expected_cols: List[str], data: pd.DataFrame):
    """Check if a list of given columns exist in a given Pandas dataframe.

    Args:
        expected_cols (List[str]): list of columns to check
        df (pd.DataFrame): pandas dataframe to check

    Raises:
        ValueError: if any of the expected columns are missing.
    """
    missing_cols = [i for i in expected_cols if i not in data.columns]
    if missing_cols:
        raise ValueError(f"Missing columns in the input dataframe: {missing_cols}.")



"""Argument Chunker Widget"""
from Orange.data import Table
from Orange.widgets import gui
from Orange.widgets.widget import Input, Output, OWWidget
from Orange.data.pandas_compat import table_from_frame, table_to_frame

from ..miner.chunker import (
    get_chunk,
    get_chunk_polarity_score,
    get_chunk_rank,
    get_chunk_topic,
    get_chunk_table,
)
from ..miner.utilities import check_columns


class OWArgChunker(OWWidget):
    """Argument chunker widget.

    This widget accepts argument data with columns of text and overall score as input and output the chunks generated from arguments, as well as a table of topic information of chunks.
    """

    name = "Argument Chunker"
    description = "Chunk arguments and do topic modeling on top of that."
    icon = "icons/OWArgChunker.svg"
    want_main_area = False

    class Inputs:
        input_data = Input("Data", Table)

    class Outputs:
        chunk_data = Output("Chunk Data", Table)
        topic_data = Output("Topic Data", Table)

    def __init__(self):
        super().__init__()
        self.df_arguments = None

        # GUI
        gui.button(self.controlArea, self, "Chunk", callback=self.chunk)

    @Inputs.input_data
    def set_input_data(self, data):
        self.df_arguments = table_to_frame(data, include_metas=True)
        expected_cols = ["argument", "score"]
        check_columns(expected_cols=expected_cols, data=self.df_arguments)

    def chunk(self):
        """Call back: chunk arguments and output chunk and topic table."""
        arguments = self.df_arguments["argument"]
        arg_scores = self.df_arguments["score"]
        chunk_arg_ids, chunks = get_chunk(docs=arguments)
        chunk_p_scores = get_chunk_polarity_score(chunks=chunks)
        chunk_topics, chunk_embeds, df_topics = get_chunk_topic(chunks=chunks)
        chunk_ranks = get_chunk_rank(arg_ids=chunk_arg_ids, embeds=chunk_embeds)

        df_chunks = get_chunk_table(
            arg_ids=chunk_arg_ids,
            chunks=chunks,
            p_scores=chunk_p_scores,
            topics=chunk_topics,
            ranks=chunk_ranks,
        )

        # # list values in dataframe will encounter hashing issue.
        # # convert to string before output.
        # object_cols = df_topics.select_dtypes(include=[object]).columns
        # df_topics[object_cols] = df_topics[object_cols].astype(str)

        table_chunks = table_from_frame(df_chunks)
        table_topics = table_from_frame(df_topics)
        self.Outputs.chunk_data.send(table_chunks)
        self.Outputs.topic_data.send(table_topics)



"""Argument processor widget"""
from Orange.data import Table
from Orange.widgets import gui
from Orange.widgets.widget import Input, Output, OWWidget
from Orange.data.pandas_compat import table_from_frame, table_to_frame

from ..miner.processor import (
    get_argument_topics,
    get_argument_coherence,
    get_argument_sentiment,
    update_argument_table,
)
from ..miner.utilities import check_columns


class OWArgProcessor(OWWidget):
    """Argument processor widget.

    Merge chunks back into arguments with the associated meta information, and compute
    other measures like readibility, sentiment, and coherence of arguments.
    """

    name = "Argument Processor"
    description = "Process input argument table and prepare it for the mining step."
    icon = "icons/OWArgProcessor.svg"
    want_main_area = False

    # GUI variables

    class Inputs:
        argument_data = Input("Argument Data", Table)
        chunk_data = Input("Chunk Data", Table)

    class Outputs:
        argument_data = Output("Argument Data", Table)

    def __init__(self):
        super().__init__()
        self.df_arguments = None
        self.df_chunks = None

        # GUI
        gui.button(self.controlArea, self, label="Process", callback=self.process)

    @Inputs.argument_data
    def set_argument_data(self, data):
        self.df_arguments = table_to_frame(data, include_metas=True)
        expected_cols = ["argument", "score"]
        check_columns(expected_cols=expected_cols, data=self.df_arguments)

    @Inputs.chunk_data
    def set_chunk_data(self, data):
        self.df_chunks = table_to_frame(data, include_metas=True)
        expected_cols = ["argument_id", "chunk", "polarity_score", "topic", "rank"]
        check_columns(expected_cols=expected_cols, data=self.df_chunks)

    def process(self):
        """Call back: merge chunks back to arguments and compute all the measures."""
        arg_topics = get_argument_topics(
            arg_ids=self.df_chunks["argument_id"], topics=self.df_chunks["topic"]
        )
        arg_sentiments = get_argument_sentiment(
            arg_ids=self.df_chunks["argument_id"],
            ranks=self.df_chunks["rank"],
            p_scores=self.df_chunks["polarity_score"],
        )
        arg_coherences = get_argument_coherence(
            scores=self.df_arguments["score"], sentiments=arg_sentiments
        )

        condition = (
            len(arg_topics)
            == len(arg_sentiments)
            == len(arg_coherences)
            == self.df_arguments.shape[0]
        )
        if not condition:
            raise ValueError(
                f"Size of the processing result not aling to the given argument table: {len(arg_topics)}, {len(arg_sentiments)}, {len(arg_coherences)}, {self.df_arguments.shape[0]}"
            )

        df_arguments_processed = update_argument_table(
            df_arguments=self.df_arguments,
            topics=arg_topics,
            sentiments=arg_sentiments,
            coherences=arg_coherences,
        )

        table_arguments = table_from_frame(df_arguments_processed)
        self.Outputs.argument_data.send(table_arguments)



"""Test the miner module."""
import pytest
import pandas as pd

from orangearg.argument.miner.miner import (
    select_by_topic,
    get_edges,
    get_edge_weights,
    get_edge_table,
    get_node_labels,
    get_node_table,
)
from .conftest import dummy_argument_selection, dummy_edge_data, df_arguments_processed


@pytest.mark.parametrize(
    "dummy_data, dummy_topic, expected_result",
    [
        (
            pd.DataFrame({"topics": [(0), (1, 2)]}),
            0,
            pd.DataFrame({"topics": [(0)], "argument_id": [0]}),
        ),
        (
            pd.DataFrame({"topics": [(), (1, 2)]}),
            0,
            pd.DataFrame({"topics": [], "argument_id": []}),
        ),
        (
            pd.DataFrame({"topics": ["(0)", "(1, 2)"]}),
            0,
            pd.DataFrame({"topics": ["(0)"], "argument_id": [0]}),
        ),
    ],
)
def test_select_by_topic(dummy_data, dummy_topic, expected_result):
    """Unit test function selection_by_topic."""
    result = select_by_topic(data=dummy_data, topic=dummy_topic)
    assert isinstance(result, pd.DataFrame)
    assert result.compare(expected_result).empty


def test_get_edges(dummy_argument_selection):
    """Unit test function get_edges."""
    expected_result = [(0, 1), (0, 2), (0, 3), (0, 4), (1, 3), (1, 4), (2, 3), (2, 4)]

    result = get_edges(data=dummy_argument_selection)

    assert result == expected_result


def test_get_edge_weights(dummy_argument_selection):
    """Unit test function get_edge_weights."""
    edges = get_edges(dummy_argument_selection)
    expected_result = [-0.2, 0.2, 0.25, -0.6, 0.45, -0.4, 0.05, -0.8]

    result = get_edge_weights(data=dummy_argument_selection, edges=edges)

    assert result == expected_result


def test_get_edge_table(dummy_argument_selection, dummy_edge_data):
    """Unit test function get_edge_table.

    Edge nr. 0, 3, 5, 7 have their source and target swapped, as their weights are negative.
    """
    edges = get_edges(dummy_argument_selection)
    weights = get_edge_weights(data=dummy_argument_selection, edges=edges)

    result = get_edge_table(edges=edges, weights=weights)

    assert result.compare(dummy_edge_data).empty


def test_get_node_labels(dummy_argument_selection, dummy_edge_data):
    """Unit test function get_node_labels.

    Argument 3 is supportive because all its attackers are attacked by some other arguments, while argument 4 is supportive because no other arguments attack it.
    """
    dummy_indices = dummy_argument_selection.index.tolist()
    dummy_sources = dummy_edge_data["source"].tolist()
    dummy_targets = dummy_edge_data["target"].tolist()
    expected_result = ["defeated", "defeated", "defeated", "supportive", "supportive"]

    result = get_node_labels(
        indices=dummy_indices, sources=dummy_sources, targets=dummy_targets
    )

    assert result == expected_result


def test_get_node_table(dummy_argument_selection, dummy_edge_data):
    """Unit test function_get_node_table."""
    dummy_indices = dummy_argument_selection.index.tolist()
    dummy_sources = dummy_edge_data["source"].tolist()
    dummy_targets = dummy_edge_data["target"].tolist()
    dummy_arg_ids = dummy_argument_selection["argument_id"].tolist()
    dummy_arguments = dummy_argument_selection["argument"].tolist()
    dummy_scores = dummy_argument_selection["score"].tolist()
    dummy_labels = get_node_labels(
        indices=dummy_indices, sources=dummy_sources, targets=dummy_targets
    )
    expected_result = pd.DataFrame(
        {
            "argument_id": [4, 7, 13, 29, 33],
            "argument": ["arg1", "arg2", "arg3", "arg4", "arg5"],
            "score": [1, 2, 2, 3, 3],
            "label": ["defeated", "defeated", "defeated", "supportive", "supportive"],
        }
    )

    result = get_node_table(
        arg_ids=dummy_arg_ids,
        arguments=dummy_arguments,
        scores=dummy_scores,
        labels=dummy_labels,
    )

    assert result.compare(expected_result).empty


def test_integrate_miner(df_arguments_processed):
    """Integrate test the miner module."""
    selection = select_by_topic(data=df_arguments_processed, topic=22)
    assert selection.shape[0] == 19

    edges = get_edges(data=selection)
    weights = get_edge_weights(data=selection, edges=edges)
    df_edges = get_edge_table(edges=edges, weights=weights)
    assert df_edges.shape[0] == 50

    indices = selection.index.tolist()
    sources = df_edges["source"].tolist()
    targets = df_edges["target"].tolist()
    labels = get_node_labels(indices=indices, sources=sources, targets=targets)
    assert labels[5] == labels[6] == labels[14] == "defeated"

    arg_ids = selection["argument_id"].tolist()
    arguments = selection["argument"].tolist()
    scores = selection["score"].tolist()
    df_nodes = get_node_table(
        arg_ids=arg_ids, arguments=arguments, scores=scores, labels=labels
    )



"""Test the utilities module."""
import pytest
import pandas as pd

from orangearg.argument.miner.utilities import check_columns


def test_check_columns():
    """Unit test check_columns."""
    dummy_df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    dummy_expected_cols = ["col1", "col3"]
    with pytest.raises(ValueError) as ex:
        check_columns(expected_cols=dummy_expected_cols, data=dummy_df)
    assert str(ex.value) == "Missing columns in the input dataframe: ['col3']."
